# encoding: utf-8
__all__ = ["CalendarFrame"]

import calendar
import locale
import tkinter as tk
import tkinter.ttk as ttk

imgp_data = (
    "R0lGODlhDAAMAIABAAAAAP///yH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEK"
    + "AAEALAAAAAAMAAwAAAIVjI+JoMsdgIRyqmoTfrfCmDWh+DUFADs="
)
imgn_data = (
    "R0lGODlhDAAMAIABAAAAAP///yH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEK"
    + "AAEALAAAAAAMAAwAAAIUjI8ZoAnczINtUmdrVpu/uFwcSBYAOw=="
)


def get_calendar(locale, fwday):
    # instantiate proper calendar class
    if locale is None:
        return calendar.TextCalendar(fwday)
    else:
        return calendar.LocaleTextCalendar(fwday, locale)


def i2rc(i, coldim):
    c = i % coldim
    f = (i - c) // coldim
    return (f, c)


def rowmajor(rows, cols):
    size = rows * cols
    for i in range(0, size):
        c = i % cols
        f = (i - c) // cols
        yield (i, f, c)


def matrix_coords(rows, cols, rowh, colw, ox=0, oy=0):
    "Generate coords for a matrix of rects"
    for i, f, c in rowmajor(rows, cols):
        x = ox + c * colw
        y = oy + f * rowh
        x1 = x + colw
        y1 = y + rowh
        yield (i, x, y, x1, y1)


class CalendarFrame(ttk.Frame):
    """Allows to choose a date in a calendar.

    WIDGET-SPECIFIC OPTIONS
            locale, firstweekday, year, month
            calendarfg, calendarbg,
            headerfg, headerbg,
            selectbg, selectfg,
            markbg, markfg,
    Generates:
        <<CalendarFrameDateSelected>>
    """

    datetime = calendar.datetime.datetime
    timedelta = calendar.datetime.timedelta

    def __init__(self, master=None, **kw):
        self.__redraw_cb = None  # For redraw callback check.
        self.__markdays_cb = None  # For markdays callback check.

        sysloc = locale.getlocale(locale.LC_TIME)
        if None in sysloc:
            sysloc = None
        else:
            sysloc = "{0}.{1}".format(*sysloc)

        self.__options = options = {
            "firstweekday": calendar.SUNDAY,
            "year": self.datetime.now().year,
            "month": self.datetime.now().month,
            "locale": sysloc,
            "calendarfg": "black",
            "calendarbg": "white",
            "headerfg": "black",
            "headerbg": "grey90",
            "selectbg": "#8000FF",
            "selectfg": "white",
            "state": "normal",
            "markbg": "white",
            "markfg": "blue",
        }
        # remove custom options from kw before initialization ttk.Frame
        for k, v in options.items():
            options[k] = kw.pop(k, v)

        # Marked days
        self._marked_days = set()
        # Calendar variables
        self._date = self.datetime(options["year"], options["month"], 1)
        self._cal = get_calendar(options["locale"], options["firstweekday"])
        self._weeks = self._cal.monthdayscalendar(
            options["year"], options["month"]
        )
        self._selection = None

        # Canvas variables
        self._rheader = None
        self._theader = [0 for x in range(0, 7)]
        self._recmat = [0 for x in rowmajor(6, 7)]
        self._txtmat = [0 for x in rowmajor(6, 7)]

        # button bar variables
        self.__img_prev = None
        self.__img_next = None
        self._lmonth = None
        self._lyear = None

        ttk.Frame.__init__(self, master, **kw)

        # build ui
        self.__build_ui()

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        color_change = False
        for key in (
            "calendarfg",
            "calendarbg",
            "headerfg",
            "headerbg",
            "selectbg",
            "selectfg",
            "markbg",
            "markfg",
        ):
            if key in kw:
                self.__options[key] = kw.pop(key)
                color_change = True
        key = "state"
        if key in kw:
            value = kw.pop(key)
            self.__options[key] = value
            self._canvas.config(state=value)
            for w in self._topframe.winfo_children():
                if w.winfo_class() == "TButton":
                    w.config(state=value)

        calendar_change = False
        key = "locale"
        if key in kw:
            value = locale.normalize(kw.pop(key))
            self.__options[key] = value
            calendar_change = True
        key = "firstweekday"
        if key in kw:
            value = kw.pop(key)
            self.__options[key] = int(value)
            calendar_change = True
        if calendar_change:
            self._reconfigure_calendar()

        date_change = False
        for key in ("year", "month"):
            if key in kw:
                self.__options[key] = int(kw.pop(key))
                date_change = True
        if date_change:
            self._reconfigure_date()

        if color_change or calendar_change or date_change:
            self._redraw_calendar()
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        if key in (
            "locale",
            "firstweekday",
            "calendarfg",
            "calendarbg",
            "headerfg",
            "headerbg",
            "selectbg",
            "selectfg",
            "markbg",
            "markfg",
            "state",
        ):
            return self.__options[key]
        option = "year"
        if key == option:
            return self._date.year
        option = "month"
        if key == option:
            return self._date.month
        return ttk.Frame.cget(self, key)

    __getitem__ = cget

    def __build_ui(self):
        # BUILD UI

        self.configure(height="200", width="200")
        self._topframe = ttk.Frame(self)
        self._topframe.configure(height="200", width="200")
        self.bpmonth = ttk.Button(self._topframe)
        self.bpmonth.configure(style="Toolbutton", text="L")
        self.bpmonth.pack(side="left")
        self.bnmonth = ttk.Button(self._topframe)
        self.bnmonth.configure(style="Toolbutton", text="R")
        self.bnmonth.pack(side="left")
        self._lmonth = ttk.Label(self._topframe)
        self._lmonth.configure(anchor="center", text="January")
        self._lmonth.pack(side="left")
        self.btoday = ttk.Button(self._topframe)
        self.btoday.configure(style="Toolbutton", text="Today")
        self.btoday.pack(expand="true", fill="x", side="left")
        self._lyear = ttk.Label(self._topframe)
        self._lyear.configure(text="2020")
        self._lyear.pack(side="left")
        self.bpyear = ttk.Button(self._topframe)
        self.bpyear.configure(style="Toolbutton", text="L")
        self.bpyear.pack(side="left")
        self.bnyear = ttk.Button(self._topframe)
        self.bnyear.configure(style="Toolbutton", text="R")
        self.bnyear.pack(side="left")
        self._topframe.pack(anchor="n", fill="x", side="top")
        self._canvas = tk.Canvas(self)
        self._canvas.configure(
            background="#ffffff",
            borderwidth="0",
            height="160",
            highlightthickness="0",
        )
        self._canvas.configure(width="240")
        self._canvas.pack(
            anchor="center", expand="true", fill="both", side="top"
        )

        self.__img_prev = imgp = tk.PhotoImage(data=imgp_data)
        self.__img_next = imgn = tk.PhotoImage(data=imgn_data)
        # self._lmonth = lmonth
        # self._lyear = lyear

        def callback(event=None):
            return self._change_date("month", -1)

        self.bpmonth.configure(image=imgp, command=callback)

        def callback(event=None):
            return self._change_date("month", 1)

        self.bnmonth.configure(image=imgn, command=callback)

        def callback(event=None):
            return self._change_date("year", -1)

        self.bpyear.configure(image=imgp, command=callback)

        def callback(event=None):
            return self._change_date("year", 1)

        self.bnyear.configure(image=imgn, command=callback)
        self.btoday.configure(command=self._go_today)
        self._canvas.bind("<Configure>", self._on_canvas_configure)
        # self._topframe = frame2
        # self._canvas = canvas
        self._draw_calendar(self._canvas)
        self._canvas.tag_bind("cell", "<Button-1>", self._on_cell_clicked)

    def _reconfigure_calendar(self):
        options = self.__options
        self._date = self.datetime(options["year"], options["month"], 1)
        self._cal = get_calendar(options["locale"], options["firstweekday"])
        self._weeks = self._cal.monthdayscalendar(
            options["year"], options["month"]
        )

    def _reconfigure_date(self):
        options = self.__options
        self._date = self.datetime(options["year"], options["month"], 1)
        self._weeks = self._cal.monthdayscalendar(
            options["year"], options["month"]
        )
        self._selection = None  # Forget current selected day
        self._redraw_calendar()

    def _go_today(self, event=None):
        options = self.__options
        today = self.datetime.now()
        options["year"] = today.year
        options["month"] = today.month
        self._reconfigure_date()

    def _change_date(self, element, direction):
        options = self.__options
        newdate = None
        if element == "month":
            if direction == -1:
                newdate = self._date - self.timedelta(days=1)
            else:
                year, month = self._date.year, self._date.month
                days = calendar.monthrange(year, month)[1] + 1
                newdate = self._date + self.timedelta(days=days)
        elif element == "year":
            year = self._date.year + direction
            newdate = self.datetime(year, self._date.month, 1)
        options["year"] = newdate.year
        options["month"] = newdate.month
        self._reconfigure_date()

    def _on_cell_clicked(self, event=None):
        item = self._canvas.find_withtag("current")
        idx = self._recmat.index(item[0])
        weeks = self._weeks
        day = 0
        f, c = i2rc(idx, 7)
        if f < len(weeks):
            day = weeks[f][c]
            if day != 0:
                self.select_day(day, self._date.month, self._date.year)

    def _mark_days(self):
        options = self.__options
        year = self._date.year
        month = self._date.month
        weeks = self._weeks
        now = self.datetime.now()
        today = (now.year, now.month, now.day)
        for i, f, c in rowmajor(6, 7):
            day = 0
            clear = True
            if f < len(weeks):
                day = weeks[f][c]
                key = (year, month, day)
                if (
                    (None, None, day) in self._marked_days
                    or (None, month, day) in self._marked_days
                    or key in self._marked_days
                ):
                    self._canvas.itemconfigure(
                        self._recmat[i],
                        fill=options["markbg"],
                        outline=options["markbg"],
                    )
                    self._canvas.itemconfigure(
                        self._txtmat[i], fill=options["markfg"]
                    )
                    clear = False
                if key == today:
                    self._canvas.itemconfigure(
                        self._recmat[i],
                        fill=options["selectfg"],
                        outline=options["selectbg"],
                    )
                    self._canvas.itemconfigure(
                        self._txtmat[i], fill=options["selectbg"]
                    )
                    clear = False
                if key == self._selection:
                    self._canvas.itemconfigure(
                        self._recmat[i],
                        fill=options["selectbg"],
                        outline=options["selectbg"],
                    )
                    self._canvas.itemconfigure(
                        self._txtmat[i], fill=options["selectfg"]
                    )
                    clear = False
            if clear:
                # clear day
                self._canvas.itemconfigure(
                    self._recmat[i],
                    fill=options["calendarbg"],
                    outline=options["calendarbg"],
                )
                self._canvas.itemconfigure(
                    self._txtmat[i], fill=options["calendarfg"]
                )
        self.__markdays_cb = None

    def _call_mark_days(self):
        if self.__markdays_cb is None:
            self.__markdays_cb = self.after_idle(self._mark_days)

    def _remark_date(self, day, month=None, year=None, highlight=True):
        key = (year, month, day)
        if highlight:
            self._marked_days.add(key)
        else:
            if key in self._marked_days:
                self._marked_days.remove(key)
        self._call_mark_days()

    def mark_day(self, day, month=None, year=None):
        """Marks the specified month day with a visual marker
        (typically by making the number bold).
        If only day is specified and the calendar month and year
        are changed, the marked day remain marked.
        You can be more specific setting month and year parameters.
        """
        self._remark_date(day, month, year, highlight=True)

    def unmark_day(self, day, month=None, year=None):
        self._remark_date(day, month, year, highlight=False)

    def clear_marks(self):
        """Clears all marked days"""
        self._marked_days.clear()
        self._call_mark_days()

    def _draw_calendar(self, canvas, redraw=False):
        """Draws calendar."""
        options = self.__options
        # Update labels:
        name = self._cal.formatmonthname(
            self._date.year, self._date.month, 0, withyear=False
        )
        self._lmonth.configure(text=name.title())
        self._lyear.configure(text=str(self._date.year))

        # Update calendar
        ch = canvas.winfo_height()
        cw = canvas.winfo_width()
        rowh = ch / 7.0
        colw = cw / 7.0
        # Header background
        if self._rheader is None:
            self._rheader = canvas.create_rectangle(
                0, 0, cw, rowh, width=0, fill=options["headerbg"]
            )
        else:
            canvas.itemconfigure(self._rheader, fill=options["headerbg"])
            canvas.coords(self._rheader, 0, 0, cw, rowh)
        # Header text
        ox = 0
        oy = rowh / 2.0
        coffset = colw / 2.0
        cols = self._cal.formatweekheader(3).split()
        for i in range(0, 7):
            x = ox + i * colw + coffset
            if redraw:
                item = self._theader[i]
                canvas.coords(item, x, oy)
                canvas.itemconfigure(
                    item, text=cols[i], fill=options["headerfg"]
                )
            else:
                self._theader[i] = canvas.create_text(
                    x, oy, text=cols[i], fill=options["headerbg"]
                )

        # background matrix
        oy = rowh
        ox = 0
        for i, x, y, x1, y1 in matrix_coords(6, 7, rowh, colw, ox, oy):
            x1 -= 1
            y1 -= 1
            if redraw:
                rec = self._recmat[i]
                canvas.coords(rec, x, y, x1, y1)
                canvas.itemconfigure(rec, fill=options["calendarbg"])
            else:
                rec = canvas.create_rectangle(
                    x,
                    y,
                    x1,
                    y1,
                    width=1,
                    fill=options["calendarbg"],
                    outline=options["calendarbg"],
                    activeoutline=options["selectbg"],
                    activewidth=1,
                    tags="cell",
                )
                self._recmat[i] = rec

        # text matrix
        weeks = self._weeks
        # xoffset = colw / 2.0
        yoffset = rowh / 2.0
        oy = rowh
        ox = 0
        for i, x, y, x1, y1 in matrix_coords(6, 7, rowh, colw, ox, oy):
            x += coffset
            y += yoffset
            # day text
            txt = ""
            f, c = i2rc(i, 7)
            if f < len(weeks):
                day = weeks[f][c]
                txt = "{0}".format(day) if day != 0 else ""
            if redraw:
                item = self._txtmat[i]
                canvas.coords(item, x, y)
                canvas.itemconfigure(item, text=txt)
            else:
                self._txtmat[i] = canvas.create_text(
                    x, y, text=txt, state=tk.DISABLED
                )
        # Mark days
        self._mark_days()

    def _redraw_calendar(self):
        self._draw_calendar(self._canvas, redraw=True)
        # after idle callback trick
        self.__redraw_cb = None

    def _on_canvas_configure(self, event=None):
        if self.__redraw_cb is None:
            self.__redraw_cb = self.after_idle(self._redraw_calendar)

    @property
    def selection(self):
        """Return a datetime representing the current selected date."""
        if not self._selection:
            return None

        year, month = self._date.year, self._date.month
        return self.datetime(year, month, self._selection[2])

    def select_day(self, day, month=None, year=None):
        options = self.__options
        options["month"] = month = self._date.month if month is None else month
        options["year"] = year = self._date.year if year is None else year
        self._reconfigure_date()
        self._selection = (year, month, day)
        self._call_mark_days()
        self.event_generate("<<CalendarFrameDateSelected>>")


if __name__ == "__main__":
    import random

    locale.setlocale(locale.LC_ALL, locale.getdefaultlocale())
    root = tk.Tk()
    c = CalendarFrame(root)
    c.grid()

    # select day
    c.select_day(1)

    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    root.mainloop()



import importlib
import pkgutil
import pygubu.plugins

from .plugin_engine import PluginRegistry, IBuilderLoaderPlugin, IDesignerPlugin


def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    #
    # Source: https://packaging.python.org/guides/creating-and-discovering-plugins/
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


class PluginManager:
    plugins = []
    designer_plugins = []

    @classmethod
    def load_plugins(cls):
        for _, name, _ in iter_namespace(pygubu.plugins):
            importlib.import_module(name)

        for class_ in PluginRegistry.plugins:
            plugin = class_()
            if plugin.do_activate():
                cls.plugins.append(plugin)

    @classmethod
    def builder_plugins(cls):
        return (
            plugin
            for plugin in cls.plugins
            if isinstance(plugin, IBuilderLoaderPlugin)
        )

    @classmethod
    def load_designer_plugins(cls):
        for plugin in cls.plugins:
            helper = plugin.get_designer_plugin()
            if helper:
                cls.designer_plugins.append(helper)

    @classmethod
    def get_preview_builder_for(cls, builder_uid: str):
        builder = None
        for plugin in cls.designer_plugins:
            builder = plugin.get_preview_builder(builder_uid)
            if builder is not None:
                break
        return builder

    @classmethod
    def get_toplevel_preview_for(
        cls, builder_uid, widget_id, builder, top_master
    ):
        top_preview = None
        for plugin in cls.designer_plugins:
            top_preview = plugin.get_toplevel_preview_for(
                builder_uid, widget_id, builder, top_master
            )
            if top_preview is not None:
                break
        return top_preview

    @classmethod
    def configure_for_preview(cls, builder_uid: str, target):
        for plugin in cls.designer_plugins:
            plugin.configure_for_preview(builder_uid, target)



from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.plugins.tk.tkstdwidgets import TKFrame as TKFrameBO
from pygubu.plugins.tk.tkstdwidgets import TKLabel as TKLabelBO
from pygubu.plugins.tk.tkstdwidgets import TKEntry as TKEntryBO
from pygubu.plugins.tk.tkstdwidgets import TKCanvas as TKCanvasBO
from customtkinter import (
    CTkFrame,
    CTkLabel,
    CTkButton,
    CTkProgressBar,
    CTkSlider,
    CTkEntry,
    CTkOptionMenu,
    CTkComboBox,
    CTkCheckBox,
    CTkRadioButton,
    CTkSwitch,
    CTkTextbox,
    CTkCanvas,
    CTkScrollbar,
)

from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import CTkBaseMixin

GCONTAINER = 0
GINPUT = 1
GDISPLAY = 2


class CTkFrameBO(CTkBaseMixin, TKFrameBO):
    class_ = CTkFrame
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
    )
    properties = (
        TKFrameBO.OPTIONS_STANDARD + TKFrameBO.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )


_builder_uid = f"{_plugin_uid}.CTkFrame"
register_widget(
    _builder_uid,
    CTkFrameBO,
    "CTkFrame",
    ("ttk", _designer_tab_label),
    group=GCONTAINER,
)


class CTkLabelBO(CTkBaseMixin, TKLabelBO):
    class_ = CTkLabel
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
    )
    properties = (
        TKLabelBO.OPTIONS_STANDARD + TKLabelBO.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )


_builder_uid = f"{_plugin_uid}.CTkLabel"
register_widget(
    _builder_uid,
    CTkLabelBO,
    "CTkLabel",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)


class CTkProgressBarBO(CTkBaseMixin, BuilderObject):
    class_ = CTkProgressBar
    allow_bindings = False
    OPTIONS_SPECIFIC = ("width", "height", "variable", "orient", "mode")
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "progress_color",
        "determinate_speed",
        "indeterminate_speed",
    )
    properties = OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ("orient",)


_builder_uid = f"{_plugin_uid}.CTkProgressBar"
register_widget(
    _builder_uid,
    CTkProgressBarBO,
    "CTkProgressBar",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)


class CTkButtonBO(CTkBaseMixin, BuilderObject):
    class_ = CTkButton
    allow_bindings = False
    OPTIONS_SPECIFIC = (
        "text",
        "width",
        "height",
        "textvariable",
        "image",
        "compound",
        "state",
        "command",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "hover_color",
        "text_color",
        "text_color_disabled",
        "text_font",
        "hover",
    )
    properties = OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ("hover",)


_builder_uid = f"{_plugin_uid}.CTkButton"
register_widget(
    _builder_uid,
    CTkButtonBO,
    "CTkButton",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkSliderBO(CTkBaseMixin, BuilderObject):
    class_ = CTkSlider
    allow_bindings = False
    OPTIONS_SPECIFIC = (
        "width",
        "height",
        "orient",
        "variable",
        "from_",
        "to",
        "command",
        "state",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "progress_color",
        "button_color",
        "button_hover_color",
        "button_corner_radius",
        "button_length",
        "number_of_steps",
    )
    properties = OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ("orient", "button_length")


_builder_uid = f"{_plugin_uid}.CTkSlider"
register_widget(
    _builder_uid,
    CTkSliderBO,
    "CTkSlider",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkEntryBO(CTkBaseMixin, TKEntryBO):
    class_ = CTkEntry
    allow_bindings = False
    OPTIONS_CUSTOM = TKEntryBO.OPTIONS_CUSTOM + (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "text_color",
        "placeholder_text_color",
        "text_font",
        "placeholder_text",
    )
    properties = (
        TKEntryBO.OPTIONS_STANDARD + TKEntryBO.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )

    def _set_property(self, target_widget, pname, value):
        real_entry = target_widget.entry
        if pname == "text":
            wstate = str(real_entry["state"])
            if wstate != "normal":
                # change state temporarily
                real_entry["state"] = "normal"
            real_entry.delete(0, "end")
            real_entry.insert(0, value)
            real_entry["state"] = wstate
        else:
            super()._set_property(target_widget, pname, value)


_builder_uid = f"{_plugin_uid}.CTkEntry"
register_widget(
    _builder_uid,
    CTkEntryBO,
    "CTkEntry",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkOptionMenuBO(CTkBaseMixin, BuilderObject):
    class_ = CTkOptionMenu
    allow_bindings = False
    OPTIONS_SPECIFIC = ("command", "variable", "values")
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "button_color",
        "button_hover_color",
        "text_color",
        "text_color_disabled",
        "dropdown_hover_color",
        "dropdown_text_color",
        "dropdown_color",
        "width",
        "height",
        "corner_radius",
        "text_font",
        "state",
        "dynamic_resizing",
    )
    properties = (
        BuilderObject.OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    command_properties = ("command",)

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("current_value",)


_builder_uid = f"{_plugin_uid}.CTkOptionMenu"
register_widget(
    _builder_uid,
    CTkOptionMenuBO,
    "CTkOptionMenu",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkComboBoxBO(CTkBaseMixin, BuilderObject):
    class_ = CTkComboBox
    allow_bindings = False
    OPTIONS_SPECIFIC = ("command", "variable", "values")
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "button_color",
        "button_hover_color",
        "text_color",
        "text_color_disabled",
        "dropdown_hover_color",
        "dropdown_text_color",
        "dropdown_color",
        "width",
        "height",
        "corner_radius",
        "text_font",
        "dropdown_text_font",
        "state",
        "hover",
    )
    properties = (
        BuilderObject.OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    command_properties = ("command",)
    ro_properties = ("hover",)

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("value",)


_builder_uid = f"{_plugin_uid}.CTkComboBox"
register_widget(
    _builder_uid,
    CTkComboBoxBO,
    "CTkComboBox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkCheckBoxBO(CTkBaseMixin, BuilderObject):
    class_ = CTkCheckBox
    allow_bindings = False
    OPTIONS_STANDARD = ("textvariable",)
    OPTIONS_SPECIFIC = (
        "command",
        "variable",
        "variable",
        "onvalue",
        "offvalue",
        "state",
        "width",
        "height",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "checkmark_color",
        "text",
        "text_font",
        "text_color",
        "text_color_disabled",
        "corner_radius",
        "hover",
        "hover_color",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("command",)
    ro_properties = ("hover",)


_builder_uid = f"{_plugin_uid}.CTkCheckBox"
register_widget(
    _builder_uid,
    CTkCheckBoxBO,
    "CTkCheckBox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkRadioButtonBO(CTkBaseMixin, BuilderObject):
    class_ = CTkRadioButton
    allow_bindings = False
    OPTIONS_STANDARD = ("textvariable",)
    OPTIONS_SPECIFIC = (
        "command",
        "variable",
        "variable",
        "onvalue",
        "offvalue",
        "state",
        "width",
        "height",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "checkmark_color",
        "text",
        "text_font",
        "text_color",
        "text_color_disabled",
        "corner_radius",
        "hover",
        "hover_color",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("command",)
    ro_properties = ("hover",)


_builder_uid = f"{_plugin_uid}.CTkRadioButton"
register_widget(
    _builder_uid,
    CTkRadioButtonBO,
    "CTkRadioButton",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkSwitchBO(CTkBaseMixin, BuilderObject):
    class_ = CTkSwitch
    allow_bindings = False
    OPTIONS_STANDARD = ("textvariable",)
    OPTIONS_SPECIFIC = (
        "command",
        "variable",
        "variable",
        "onvalue",
        "offvalue",
        "state",
        "width",
        "height",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "checkmark_color",
        "text",
        "text_font",
        "text_color",
        "text_color_disabled",
        "corner_radius",
        "hover",
        "hover_color",
        "progress_color",
        "button_color",
        "button_hover_color",
        "button_length",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("command",)
    ro_properties = ("hover", "text_color", "text_color_disabled")


_builder_uid = f"{_plugin_uid}.CTkSwitch"
register_widget(
    _builder_uid,
    CTkSwitchBO,
    "CTkSwitch",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkTextboxBO(CTkBaseMixin, BuilderObject):
    class_ = CTkTextbox
    OPTIONS_SPECIFIC = (
        "width",
        "height",
    )
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "text_font",
        "text_color",
    )
    properties = OPTIONS_SPECIFIC + OPTIONS_CUSTOM


_builder_uid = f"{_plugin_uid}.CTkTextbox"
register_widget(
    _builder_uid,
    CTkTextboxBO,
    "CTkTextbox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkCanvasBO(CTkBaseMixin, TKCanvasBO):
    class_ = CTkCanvas


_builder_uid = f"{_plugin_uid}.CTkCanvas"
register_widget(
    _builder_uid,
    CTkCanvasBO,
    "CTkCanvas",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)


class CTkScrollbarBO(CTkBaseMixin, BuilderObject):
    class_ = CTkScrollbar
    OPTIONS_SPECIFIC = ("width", "height", "orientation", "command")
    OPTIONS_CUSTOM = (
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "scrollbar_color",
        "scrollbar_hover_color",
        "border_spacing",
        "minimum_pixel_length",
        "hover",
    )
    properties = OPTIONS_SPECIFIC + OPTIONS_CUSTOM


_builder_uid = f"{_plugin_uid}.CTkScrollbar"
register_widget(
    _builder_uid,
    CTkScrollbarBO,
    "CTkScrollbar",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)



# encoding: UTF-8
import xml.etree.ElementTree as ET
from collections import namedtuple

from .builderobject import CLASS_MAP

__all__ = ["WidgetMeta", "BindingMeta"]

BindingMeta = namedtuple("BindingMeta", ["sequence", "handler", "add"])

GridRCLine = namedtuple("GridRCLine", ["rctype", "rcid", "pname", "pvalue"])


class WidgetMeta:
    def __init__(
        self,
        cname,
        identifier,
        manager=None,
        properties_defaults=None,
        layout_defaults=None,
    ):
        super().__init__()
        self.classname = cname
        self._id = identifier
        self.properties = {}
        self.bindings = []
        self._manager = manager if manager is not None else "grid"
        widget_description = CLASS_MAP.get(cname)
        if widget_description:
            self.layout_required = widget_description.builder.layout_required
        else:
            self.layout_required = True
        self.layout_properties = {}
        self._container_manager = self._manager
        self.container_properties = {}
        self.gridrc_properties = []
        self.properties_defaults = (
            properties_defaults if properties_defaults is not None else {}
        )
        self.layout_defaults = (
            layout_defaults if layout_defaults is not None else {}
        )
        self._named = False
        # init defaults
        self.apply_properties_defaults()
        self.apply_layout_defaults()

    @property
    def manager(self):
        return self._manager

    @manager.setter
    def manager(self, value):
        self._manager = value

    @property
    def container_manager(self):
        return self._container_manager

    @container_manager.setter
    def container_manager(self, value):
        if value == "pack" and self.gridrc_properties:
            self.gridrc_properties.clear()
        self._container_manager = value

    @property
    def is_named(self) -> bool:
        return self._named

    @is_named.setter
    def is_named(self, value: bool):
        self._named = value

    @property
    def identifier(self) -> str:
        return self._id

    @identifier.setter
    def identifier(self, value: str):
        if not value:
            raise ValueError()
        self._id = value

    def apply_properties_defaults(self):
        for name, value in self.properties_defaults.items():
            self.properties[name] = value

    def apply_layout_defaults(self):
        layout_defaults_by_manager = self.layout_defaults
        if self.manager in self.layout_defaults:
            layout_defaults_by_manager = self.layout_defaults[self.manager]
        for name, value in layout_defaults_by_manager.items():
            self.layout_properties[name] = value

    def has_layout_defined(self):
        return len(self.layout_properties) > 0

    def clear_layout(self):
        self.layout_properties = {}
        self.gridrc_properties = []
        self.apply_layout_defaults()

    def get_gridrc_value(self, rctype, rcid, pname):
        value = None
        for line in self.gridrc_properties:
            if (
                line.rctype == rctype
                and line.rcid == rcid
                and line.pname == pname
            ):
                value = line.pvalue
                break
        return value

    def set_gridrc_value(self, rctype, rcid, pname, value):
        index = None
        for i, r in enumerate(self.gridrc_properties):
            if r.rctype == rctype and r.rcid == rcid and r.pname == pname:
                index = i
                break
        if index is None:
            # We're setting the grid rc property on this widget for the first time.

            line = GridRCLine(rctype, rcid, pname, value)
            self.gridrc_properties.append(line)
        else:
            # We're updating an existing grid rc property value.

            # Prevent code such as weight='0', uniform='' from showing up
            # in the generated code - it would be redundant.
            if (pname in ("minsize", "pad", "weight") and value == "0") or (
                pname == "uniform" and not value
            ):

                # We found a redundant value
                # '0' or a blank string if it's for the property: uniform

                # Remove the gridrc property
                self.gridrc_properties.pop(index)
            else:
                # Update the gridrc property
                line = GridRCLine(rctype, rcid, pname, value)
                self.gridrc_properties[index] = line

    def __repr__(self):
        tpl = """<WidgetMeta classname: {0} identifier: {1}>"""
        return tpl.format(self.classname, self.identifier)

    def copy_gridrc(self, from_, rctype):
        """Copy gridrc lines of type rctype from from_"""
        rc = [line for line in self.gridrc_properties if line.rctype != rctype]
        for line in from_.gridrc_properties:
            if line.rctype == rctype:
                rc.append(line)
        self.gridrc_properties = rc

    def copy_properties(self, wfrom):
        # Used on preview methods
        self.properties = wfrom.properties.copy()
        self.gridrc_properties.clear()
        self.gridrc_properties.extend(wfrom.gridrc_properties)
        self.container_manager = wfrom.container_manager
        self.container_properties = wfrom.container_properties


if __name__ == "__main__":
    w = WidgetMeta("mywidget", "w1")
    w.manager = "pack"
    w.properties = {"key1": "value1", "key2": "value2"}
    w.bindings = [
        BindingMeta("<<event_1>>", "callback_1", ""),
        BindingMeta("<<event_2>>", "callback_2", "+"),
    ]
    w.layout_properties = {"prop1": "value1", "prop2": "value2"}

    print("To xml")
    node = w.to_xmlnode()
    node = ET.tostring(node)
    print(node, end="\n\n")

    print("From xml:")
    strdata = """<?xml version='1.0' encoding='utf-8'?>
<object class="ttk.Label" id="Label_1">
        <property name="text" translatable="yes">Label_1</property>
        <property name="background">yellow</property>
        <layout manager="grid">
          <property name="propagate">True</property>
          <property type="row" id="0" name="pad">20</property>
          <property type="col" id="0" name="pad">10</property>
          <property type="row" id="0" name="weight">1</property>
        </layout>
      </object>
    """
    node = ET.fromstring(strdata)
    meta = WidgetMeta.from_xmlnode(node)
    print(meta)



import tkinter as tk
from customtkinter import CTkFrame, set_appearance_mode, set_default_color_theme
from pygubu.api.v1 import BuilderObject
from pygubu.plugins.pygubu.designer.basehelpers import (
    ToplevelPreviewBaseBO,
    ToplevelPreviewFactory,
    ToplevelPreviewMixin,
)
from ..widgets import CTkFrameBO


# CTkFrame has a hidden canvas inside. So, to make it clickable on preview
# we need a hack.


class CTkFrameForPreview(CTkFrame):
    def winfo_children(self):
        return super(tk.Frame, self).winfo_children()


#
# Preview classes for CTKToplevel
#
CTKToplevelPreview = ToplevelPreviewFactory(
    "CTKToplevelPreview",
    (ToplevelPreviewMixin, CTkFrameForPreview, object),
    {},
)


class CTkToplevelPreviewBO(ToplevelPreviewBaseBO):
    class_ = CTKToplevelPreview
    ro_properties = ToplevelPreviewBaseBO.ro_properties + ("background",)


#
# Preview classes for CTK
#
CTKPreview = ToplevelPreviewFactory(
    "CTKPreview",
    (ToplevelPreviewMixin, CTkFrameForPreview, object),
    {},
)


class CTkPreviewBO(ToplevelPreviewBaseBO):
    class_ = CTKPreview
    properties = ToplevelPreviewBaseBO.properties + ("appearance_mode",)
    ro_properties = ToplevelPreviewBaseBO.ro_properties + ("background",)

    def _set_property(self, target_widget, pname, value):
        if pname == "appearance_mode":
            set_appearance_mode(value)
        elif pname == "color_theme":
            set_default_color_theme(value)
        else:
            return super()._set_property(target_widget, pname, value)


class CTkFramePreviewBO(CTkFrameBO):
    class_ = CTkFrameForPreview



import tkinter as tk
from datetime import datetime
from pygubu.api.v1 import BuilderObject
from pygubu.i18n import _


class CalendarBaseBO(BuilderObject):
    OPTIONS_CUSTOM = (
        "year",
        "month",
        "day",
        "firstweekday",
        "mindate",
        "maxdate",
        "showweeknumbers",
        "showothermonthdays",
        "date_pattern",
        "selectmode",
        "textvariable",
        "background",
        "foreground",
        "disabledbackground",
        "disabledforeground",
        "bordercolor",
        "headersbackground",
        "headersforeground",
        "selectbackground",
        "selectforeground",
        "disabledselectbackground",
        "disabledselectforeground",
        "normalbackground",
        "normalforeground",
        "weekendbackground",
        "weekendforeground",
        "othermonthforeground",
        "othermonthbackground",
        "othermonthweforeground",
        "othermonthwebackground",
        "disableddaybackground",
        "disableddayforeground",
        "tooltipforeground",
        "tooltipbackground",
        "tooltipalpha",
        "tooltipdelay",
    )
    ro_properties = ("year", "month", "day")

    def _process_property_value(self, pname, value):
        if pname in ("year", "month", "day", "tooltipdelay"):
            return int(value)
        if pname in ("showweeknumbers", "showothermonthdays"):
            return tk.getboolean(value)
        if pname == "tooltipalpha":
            return float(value)
        if pname in ("mindate", "maxdate"):
            return datetime.strptime(value, "%Y-%m-%d")
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in ("showweeknumbers", "showothermonthdays"):
            return self._process_property_value(pname, value)
        if pname in ("mindate", "maxdate"):
            return f"""Calendar.datetime.strptime("{value}", "%Y-%m-%d")"""
        return super()._code_process_property_value(targetid, pname, value)


_none = {}
_base_prop_desc = (
    (
        "year",
        "naturalnumber",
        {"help": _("intinitially displayed year, default is current year.")},
    ),
    (
        "month",
        "naturalnumber",
        {"help": _("initially displayed month, default is current month.")},
    ),
    ("day", "naturalnumber", {"help": "initially selected day"}),
    (
        "firstweekday",
        "choice",
        {
            "values": ("", "monday", "sunday"),
            "state": "readonly",
            "help": _("first day of the week"),
        },
    ),
    (
        "mindate",
        "entry",
        {"help": _("minimum allowed date (string with format yyyy-mm-dd)")},
    ),
    (
        "maxdate",
        "entry",
        {"help": _("maximum allowed date (string with format yyyy-mm-dd)")},
    ),
    (
        "showweeknumbers",
        "choice",
        {
            "values": ("", "true", "false"),
            "state": "readonly",
            "help": _("whether to display week numbers"),
        },
    ),
    (
        "showothermonthdays",
        "choice",
        {
            "values": ("", "true", "false"),
            "state": "readonly",
            "help": _(
                "whether to display the last days of the previous month and the first of the next month."
            ),
        },
    ),
    (
        "date_pattern",
        "choice",
        {
            "values": (
                "",
                "dd/mm/yy",
                "dd/mm/yyyy",
                "mm/dd/yy",
                "mm/dd/yyyy",
                "yyyy/mm/dd",
            ),
            "state": "normal",
            "help": _("date pattern used to format the date as a string."),
        },
    ),
    (
        "selectmode",
        "choice",
        {"values": ("", "none", "day"), "state": "readonly"},
    ),
    (
        "textvariable",
        "tkvarentry",
        {"help": _("connect the currently selected date to the variable.")},
    ),
    ("background", "colorentry", _none),
    ("foreground", "colorentry", _none),
    ("disabledbackground", "colorentry", _none),
    ("disabledforeground", "colorentry", _none),
    ("bordercolor", "colorentry", _none),
    ("headersbackground", "colorentry", _none),
    ("headersforeground", "colorentry", _none),
    ("selectbackground", "colorentry", _none),
    ("selectforeground", "colorentry", _none),
    ("disabledselectbackground", "colorentry", _none),
    ("disabledselectforeground", "colorentry", _none),
    ("normalbackground", "colorentry", _none),
    ("normalforeground", "colorentry", _none),
    ("weekendbackground", "colorentry", _none),
    ("weekendforeground", "colorentry", _none),
    ("othermonthforeground", "colorentry", _none),
    ("othermonthbackground", "colorentry", _none),
    ("othermonthweforeground", "colorentry", _none),
    ("othermonthwebackground", "colorentry", _none),
    ("disableddaybackground", "colorentry", _none),
    ("disableddayforeground", "colorentry", _none),
    ("tooltipforeground", "colorentry", _none),
    ("tooltipbackground", "colorentry", _none),
    (
        "tooltipalpha",
        "spinbox",
        {
            "from_": 0.1,
            "to": 1,
            "increment": 0.1,
            "help": _("tooltip opacity between 0 and 1"),
        },
    ),
    (
        "tooltipdelay",
        "naturalnumber",
        {"help": _("delay in ms before displaying the tooltip")},
    ),
)



# encoding: utf-8
from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import (
    TTKTreeviewBO,
    TTKTreeviewColumnBO,
)
from .scrollbarhelper import TTKSBHelperBO
from pygubu.i18n import _
from pygubu.widgets.editabletreeview import EditableTreeview


class EditableTreeviewBO(TTKTreeviewBO):
    class_ = EditableTreeview
    virtual_events = TTKTreeviewBO.virtual_events + (
        "<<TreeviewInplaceEdit>>",
        "<<TreeviewCellEdited>>",
    )


classid = "pygubu.builder.widgets.editabletreeview"
if classid not in TTKTreeviewColumnBO.allowed_parents:
    TTKTreeviewColumnBO.add_allowed_parent(classid)
if classid not in TTKSBHelperBO.allowed_children:
    TTKSBHelperBO.add_allowed_child(classid)

register_widget(
    "pygubu.builder.widgets.editabletreeview",
    EditableTreeviewBO,
    "EditableTreeview",
    (_("Pygubu Widgets"), "ttk"),
)
